"""
This file contains coordinate transformation methods that are very thin wrappers
of palpy methods, or that have no dependence on palpy at all
"""

__all__ = (
    "alt_az_pa_from_ra_dec",
    "_alt_az_pa_from_ra_dec",
    "spherical_from_cartesian",
    "cartesian_from_spherical",
    "xyz_from_ra_dec",
    "_xyz_from_ra_dec",
    "_ra_dec_from_xyz",
    "ra_dec_from_xyz",
    "xyz_angular_radius",
    "_xyz_angular_radius",
    "rotation_matrix_from_vectors",
    "rot_about_z",
    "rot_about_y",
    "rot_about_x",
    "calc_lmst",
    "calc_lmst_astropy",
    "angular_separation",
    "_angular_separation",
    "haversine",
    "arcsec_from_radians",
    "radians_from_arcsec",
    "arcsec_from_degrees",
    "degrees_from_arcsec",
)

import numbers

import numpy as np
from astropy import units as u
from astropy.coordinates import AltAz, EarthLocation, SkyCoord
from astropy.time import Time
from astropy.utils.iers import conf

from rubin_scheduler.utils.code_utilities import _validate_inputs


def _wrap_hour_angle(ha_rad):
    """wrap hour angle to go between -pi and pi"""

    if np.size(ha_rad) == 1:
        if ha_rad > np.pi:
            ha_rad -= 2.0 * np.pi
    else:
        over = np.where(ha_rad > np.pi)[0]
        ha_rad[over] -= 2.0 * np.pi

    return ha_rad


def alt_az_pa_from_ra_dec(ra, dec, mjd, site_longitude, site_latitude):
    """ """

    alt, az, pa = _alt_az_pa_from_ra_dec(
        np.radians(ra), np.radians(dec), mjd, np.radians(site_longitude), np.radians(site_latitude)
    )

    return np.degrees(alt), np.degrees(az), np.degrees(pa)


def _alt_az_pa_from_ra_dec(ra, dec, mjd, site_longitude, site_latitude):
    """ """

    observing_location = EarthLocation(
        lat=site_latitude * u.rad, lon=site_longitude * u.rad, height=100 * u.m
    )
    observing_time = Time(mjd, format="mjd", location=observing_location)
    aa = AltAz(location=observing_location, obstime=observing_time)
    coord = SkyCoord(ra * u.rad, dec * u.rad)
    altaz = coord.transform_to(aa)

    lmst = observing_time.sidereal_time("mean")

    hour_angle = _wrap_hour_angle(lmst.rad - ra)

    # Position Angle Equation from:
    # http://www.gb.nrao.edu/~rcreager/GBTMetrology/140ft/l0058/gbtmemo52/memo52.html
    # or
    # http://www.gb.nrao.edu/GBT/DA/gbtidl/release2pt9/contrib/contrib/parangle.pro
    pa = np.arctan2(
        np.sin(hour_angle),
        (np.cos(dec) * np.tan(site_latitude) - np.sin(dec) * np.cos(hour_angle)),
    )

    return altaz.alt.rad, altaz.az.rad, pa


def calc_lmst(mjd, longitude_rad):
    """Calculate the LMST for a location
    based on:  https://github.com/jhaupt/Sidereal-Time-Calculator/blob/master/SiderealTimeCalculator.py
    which uses:
    http://aa.usno.navy.mil/faq/docs/JD_Formula.php
    http://aa.usno.navy.mil/faq/docs/GAST.php and

    Parameters
    ----------
    mjd : `float`
        is the universal time (ut1) expressed as an MJD.
        This can be a numpy array or a single value.
    long_rad : `float`
        is the longitude in radians (positive east of the prime meridian)
        This can be numpy array or a single value.  If a numpy array,
        should have the same length as mjd.  In that
        case, each long_rad will be applied only to the corresponding mjd.

    Returns
    -------
    lst : `float`
        The local sidereal time in hours

    """
    gmst = 18.697374558 + 24.06570982441908 * (mjd - 51544.5)
    gmst = gmst % 24  # to hours
    longitude = np.degrees(longitude_rad) / 15.0  # Convert longitude to hours
    lst = gmst + longitude  # Fraction LST. If negative we want to add 24...
    if np.size(lst) == 1:
        if lst < 0:
            lst += 24
    else:
        lst[np.where(lst < 0)] += 24
    return lst


def calc_lmst_astropy(mjd, long_rad):
    """
    calculates local mean sidereal time

    Parameters
    ----------
    mjd : `float`
        is the universal time (ut1) expressed as an MJD.
        This can be a numpy array or a single value.
    long_rad : `float`
        is the longitude in radians (positive east of the prime meridian)
        This can be numpy array or a single value.  If a numpy array,
        should have the same length as mjd.  In that
        case, each long_rad will be applied only to the corresponding mjd.

    Returns
    -------
    lmst : `float`
        is the local mean sidereal time in hours
    """
    observing_location = EarthLocation(lat=0.0, lon=long_rad * u.rad, height=100 * u.m)
    t = Time(mjd, format="mjd", location=observing_location)
    # Ignore astropy if we wander too far into the future
    with conf.set_temp("iers_degraded_accuracy", "warn"):
        lmst = t.sidereal_time("mean").rad

    # convert to hours
    lmst = lmst * 12 / np.pi
    return lmst


def cartesian_from_spherical(longitude, latitude):
    """
    Transforms between spherical and Cartesian coordinates.

    Parameters
    ----------
    longitude : `Unknown`
        is a numpy array or a number in radians
    latitude : `Unknown`
        is a numpy array or number in radians
    a : `Unknown`
        numpy array of the (three-dimensional) cartesian coordinates on a unit sphere.

    if inputs are numpy arrays:
    output[i][0] will be the x-coordinate of the ith point
    output[i][1] will be the y-coordinate of the ith point
    output[i][2] will be the z-coordinate of the ith point

    All angles are in radians

    Also, look at xyz_from_ra_dec().
    """
    return _xyz_from_ra_dec(longitude, latitude).transpose()


def spherical_from_cartesian(xyz):
    """
    Transforms between Cartesian and spherical coordinates

    Parameters
    ----------
    xyz : `Unknown`
        is a numpy array of points in 3-D space.
        Each row is a different point.
    returns : `Unknown`
        longitude and latitude

    All angles are in radians

    Also, look at ra_dec_from_xyz().
    """
    if not isinstance(xyz, np.ndarray):
        raise RuntimeError("You need to pass a numpy array to spherical_from_cartesian")

    if len(xyz.shape) > 1:
        return _ra_dec_from_xyz(xyz[:, 0], xyz[:, 1], xyz[:, 2])
    else:
        return _ra_dec_from_xyz(xyz[0], xyz[1], xyz[2])


def xyz_from_ra_dec(ra, dec):
    """
    Utility to convert RA,dec positions in x,y,z space.

    Parameters
    ----------
    ra : float or array
        RA in degrees
    dec : float or array
        Dec in degrees

    Returns
    -------
    x,y,z : floats or arrays
        The position of the given points on the unit sphere.
    """
    return _xyz_from_ra_dec(np.radians(ra), np.radians(dec))


def _xyz_from_ra_dec(ra, dec):
    """
    Utility to convert RA,dec positions in x,y,z space.

    Parameters
    ----------
    ra : float or array
        RA in radians
    dec : float or array
        Dec in radians

    Returns
    -------
    x,y,z : floats or arrays
        The position of the given points on the unit sphere.
    """
    # It is ok to mix floats and numpy arrays.

    cos_dec = np.cos(dec)
    return np.array([np.cos(ra) * cos_dec, np.sin(ra) * cos_dec, np.sin(dec)])


def _ra_dec_from_xyz(x, y, z):
    """
    Utility to convert x,y,z Cartesian coordinates to RA, dec positions in space.

    Parameters
    ----------
    x : float or array
        The position on the x-axis of the given points on the unit sphere
    y : float or array
        The position on the y-axis of the given points on the unit sphere
    z : float or array
        The position on the z-axis of the given points on the unit sphere

    Returns
    -------
    ra, dec : floats or arrays
        Ra and dec coordinates in radians.
    """
    rad = np.sqrt(x**2 + y**2 + z**2)
    ra = np.arctan2(y, x)
    dec = np.arcsin(z / rad)

    return ra, dec


def ra_dec_from_xyz(x, y, z):
    """
    Utility to convert x,y,z Cartesian coordinates to RA, dec positions in space.

    Parameters
    ----------
    x : float or array
        The position on the x-axis of the given points on the unit sphere
    y : float or array
        The position on the y-axis of the given points on the unit sphere
    z : float or array
        The position on the z-axis of the given points on the unit sphere

    Returns
    -------
    ra, dec : floats or arrays
        Ra and dec coordinates in degrees.
    """

    return np.degrees(_ra_dec_from_xyz(x, y, z))


def xyz_angular_radius(radius=1.75):
    """
    Convert an angular radius into a physical radius for a kdtree search.

    Parameters
    ----------
    radius : float
        Radius in degrees.

    Returns
    -------
    radius : float
    """
    return _xyz_angular_radius(np.radians(radius))


def _xyz_angular_radius(radius):
    """
    Convert an angular radius into a physical radius for a kdtree search.

    Parameters
    ----------
    radius : float
        Radius in radians.

    Returns
    -------
    radius : float
    """
    x0, y0, z0 = (1, 0, 0)
    x1, y1, z1 = _xyz_from_ra_dec(radius, 0)
    result = np.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2 + (z1 - z0) ** 2)
    return result


def z_rotation_matrix(theta):
    cc = np.cos(theta)
    ss = np.sin(theta)
    return np.array([[cc, -ss, 0.0], [ss, cc, 0.0], [0.0, 0.0, 1.0]])


def rot_about_z(vec, theta):
    """
    Rotate a Cartesian vector an angle theta about the z axis.
    Theta is in radians.
    Positive theta rotates +x towards +y.
    """
    return np.dot(z_rotation_matrix(theta), vec.transpose()).transpose()


def y_rotation_matrix(theta):
    cc = np.cos(theta)
    ss = np.sin(theta)
    return np.array([[cc, 0.0, ss], [0.0, 1.0, 0.0], [-ss, 0.0, cc]])


def rot_about_y(vec, theta):
    """
    Rotate a Cartesian vector an angle theta about the y axis.
    Theta is in radians.
    Positive theta rotates +x towards -z.
    """
    return np.dot(y_rotation_matrix(theta), vec.transpose()).transpose()


def x_rotation_matrix(theta):
    cc = np.cos(theta)
    ss = np.sin(theta)

    return np.array([[1.0, 0.0, 0.0], [0.0, cc, -ss], [0.0, ss, cc]])


def rot_about_x(vec, theta):
    """
    Rotate a Cartesian vector an angle theta about the x axis.
    Theta is in radians.
    Positive theta rotates +y towards +z.
    """
    return np.dot(x_rotation_matrix(theta), vec.transpose()).transpose()


def rotation_matrix_from_vectors(v1, v2):
    """
    Given two vectors v1,v2 calculate the rotation matrix for v1->v2 using the axis-angle approach

    Parameters
    ----------
    v1,v2 : `Unknown`
        Cartesian unit vectors (in three dimensions).
    rot : `Unknown`
        is the rotation matrix that rotates from one to the other
    """

    if np.abs(np.sqrt(np.dot(v1, v1)) - 1.0) > 0.01:
        raise RuntimeError("v1 in rotation_matrix_from_vectors is not a unit vector")

    if np.abs(np.sqrt(np.dot(v2, v2)) - 1.0) > 0.01:
        raise RuntimeError("v2 in rotation_matrix_from_vectors is not a unit vector")

    # Calculate the axis of rotation by the cross product of v1 and v2
    cross = np.cross(v1, v2)
    cross = cross / np.sqrt(np.dot(cross, cross))

    # calculate the angle of rotation via dot product
    angle = np.arccos(np.dot(v1, v2))
    sin_dot = np.sin(angle)
    cos_dot = np.cos(angle)

    # calculate the corresponding rotation matrix
    # http://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
    rot = [
        [
            cos_dot + cross[0] * cross[0] * (1 - cos_dot),
            -cross[2] * sin_dot + (1 - cos_dot) * cross[0] * cross[1],
            cross[1] * sin_dot + (1 - cos_dot) * cross[0] * cross[2],
        ],
        [
            cross[2] * sin_dot + (1 - cos_dot) * cross[0] * cross[1],
            cos_dot + (1 - cos_dot) * cross[1] * cross[1],
            -cross[0] * sin_dot + (1 - cos_dot) * cross[1] * cross[2],
        ],
        [
            -cross[1] * sin_dot + (1 - cos_dot) * cross[0] * cross[2],
            cross[0] * sin_dot + (1 - cos_dot) * cross[1] * cross[2],
            cos_dot + (1 - cos_dot) * (cross[2] * cross[2]),
        ],
    ]

    return rot


def _angular_separation(long1, lat1, long2, lat2):
    """
    Angular separation between two points in radians

    Parameters
    ----------
    long1 is the first longitudinal coordinate in radians

    lat1 is the first latitudinal coordinate in radians

    long2 is the second longitudinal coordinate in radians

    lat2 is the second latitudinal coordinate in radians

    Returns
    -------
    The angular separation between the two points in radians

    Calculated based on the haversine formula
    From http://en.wikipedia.org/wiki/Haversine_formula
    """
    are_arrays_1 = _validate_inputs([long1, lat1], ["long1", "lat1"], "angular_separation")

    are_arrays_2 = _validate_inputs([long2, lat2], ["long2", "lat2"], "angular_separation")

    # The code below is necessary because the call to np.radians() in
    # angular_separation() will automatically convert floats
    # into length 1 numpy arrays, confusing validate_inputs()
    if are_arrays_1 and len(long1) == 1:
        are_arrays_1 = False
        long1 = long1[0]
        lat1 = lat1[0]

    if are_arrays_2 and len(long2) == 1:
        are_arrays_2 = False
        long2 = long2[0]
        lat2 = lat2[0]

    if are_arrays_1 and are_arrays_2:
        if len(long1) != len(long2):
            raise RuntimeError(
                "You need to pass arrays of the same length " "as arguments to angular_separation()"
            )

    t1 = np.sin(lat2 / 2.0 - lat1 / 2.0) ** 2
    t2 = np.cos(lat1) * np.cos(lat2) * np.sin(long2 / 2.0 - long1 / 2.0) ** 2
    _sum = t1 + t2

    if isinstance(_sum, numbers.Number):
        if _sum < 0.0:
            _sum = 0.0
    else:
        _sum = np.where(_sum < 0.0, 0.0, _sum)

    return 2.0 * np.arcsin(np.sqrt(_sum))


def angular_separation(long1, lat1, long2, lat2):
    """
    Angular separation between two points in degrees

    Parameters
    ----------
    long1 is the first longitudinal coordinate in degrees

    lat1 is the first latitudinal coordinate in degrees

    long2 is the second longitudinal coordinate in degrees

    lat2 is the second latitudinal coordinate in degrees

    Returns
    -------
    The angular separation between the two points in degrees
    """
    return np.degrees(
        _angular_separation(np.radians(long1), np.radians(lat1), np.radians(long2), np.radians(lat2))
    )


def haversine(long1, lat1, long2, lat2):
    """
    DEPRECATED; use angular_separation() instead

    Return the angular distance between two points in radians

    Parameters
    ----------
    long1 : `Unknown`
        is the longitude of point 1 in radians
    lat1 : `Unknown`
        is the latitude of point 1 in radians
    long2 : `Unknown`
        is the longitude of point 2 in radians
    lat2 : `Unknown`
        is the latitude of point 2 in radians
    the : `Unknown`
        angular separation between points 1 and 2 in radians
    """
    return _angular_separation(long1, lat1, long2, lat2)


def arcsec_from_radians(value):
    """
    Convert an angle in radians to arcseconds

    Note: if you input None, you will get None back
    """
    if value is None:
        return None

    return 3600.0 * np.degrees(value)


def radians_from_arcsec(value):
    """
    Convert an angle in arcseconds to radians

    Note: if you input None, you will get None back
    """
    if value is None:
        return None

    return np.radians(value / 3600.0)


def arcsec_from_degrees(value):
    """
    Convert an angle in degrees to arcseconds

    Note: if you input None, you will get None back
    """
    if value is None:
        return None

    return 3600.0 * value


def degrees_from_arcsec(value):
    """
    Convert an angle in arcseconds to degrees

    Note: if you input None, you will get None back
    """
    if value is None:
        return None

    return value / 3600.0



import os
import unittest

import numpy as np

import rubin_scheduler.scheduler.basis_functions as bf
import rubin_scheduler.scheduler.detailers as detailers
import rubin_scheduler.utils as utils
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.scheduler import sim_runner
from rubin_scheduler.scheduler.example import example_scheduler
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.schedulers import CoreScheduler
from rubin_scheduler.scheduler.surveys import (
    BlobSurvey,
    GreedySurvey,
    ScriptedSurvey,
    generate_ddf_scheduled_obs,
)
from rubin_scheduler.scheduler.utils import SkyAreaGenerator, calc_norm_factor_array

SAMPLE_BIG_DATA_FILE = os.path.join(get_data_dir(), "scheduler/dust_maps/dust_nside_32.npz")


class ModelObservatoryWindy(ModelObservatory):
    """Have the model observatory always have a strong wind from the north"""

    def return_conditions(self):
        """
        Returns
        -------
        rubin_scheduler.scheduler.features.conditions object
        """
        _conditions = super().return_conditions()

        # Always have a strong wind from the north
        wind_speed = 40.0
        wind_direction = 0.0
        self.conditions.wind_speed = wind_speed
        self.conditions.wind_direction = wind_direction

        return self.conditions


def ddf_surveys(detailers=None, season_unobs_frac=0.2, euclid_detailers=None, nside=None):
    obs_array = generate_ddf_scheduled_obs(season_unobs_frac=season_unobs_frac)

    euclid_obs = np.where((obs_array["note"] == "DD:EDFS_b") | (obs_array["note"] == "DD:EDFS_a"))[0]
    all_other = np.where((obs_array["note"] != "DD:EDFS_b") & (obs_array["note"] != "DD:EDFS_a"))[0]

    survey1 = ScriptedSurvey([bf.AvoidDirectWind(nside=nside)], detailers=detailers)
    survey1.set_script(obs_array[all_other])

    survey2 = ScriptedSurvey([bf.AvoidDirectWind(nside=nside)], detailers=euclid_detailers)
    survey2.set_script(obs_array[euclid_obs])

    return [survey1, survey2]


def gen_greedy_surveys(nside):
    """
    Make a quick set of greedy surveys
    """
    sky = SkyAreaGenerator(nside=nside)
    target_map, labels = sky.return_maps()
    filters = ["g", "r", "i", "z", "y"]
    surveys = []

    for filtername in filters:
        bfs = []
        bfs.append(bf.M5DiffBasisFunction(filtername=filtername, nside=nside))
        bfs.append(
            bf.TargetMapBasisFunction(
                filtername=filtername,
                target_map=target_map[filtername],
                out_of_bounds_val=np.nan,
                nside=nside,
            )
        )
        bfs.append(bf.SlewtimeBasisFunction(filtername=filtername, nside=nside))
        bfs.append(bf.StrictFilterBasisFunction(filtername=filtername))
        # Masks, give these 0 weight
        bfs.append(bf.AvoidDirectWind(nside=nside))
        bfs.append(bf.AltAzShadowMaskBasisFunction(nside=nside, shadow_minutes=60.0, max_alt=76.0))
        bfs.append(bf.MoonAvoidanceBasisFunction(nside=nside, moon_distance=30.0))
        bfs.append(bf.CloudedOutBasisFunction())

        bfs.append(bf.FilterLoadedBasisFunction(filternames=filtername))
        bfs.append(bf.PlanetMaskBasisFunction(nside=nside))

        weights = np.array([3.0, 0.3, 3.0, 3.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        surveys.append(
            GreedySurvey(
                bfs,
                weights,
                block_size=1,
                filtername=filtername,
                dither=True,
                nside=nside,
                survey_name="greedy",
            )
        )
    return surveys


def gen_blob_surveys(nside):
    """
    make a quick set of blob surveys
    """
    sky = SkyAreaGenerator(nside=nside)
    target_map, labels = sky.return_maps()
    norm_factor = calc_norm_factor_array(target_map)

    filter1s = ["g"]  # , 'r', 'i', 'z', 'y']
    filter2s = ["g"]  # , 'r', 'i', None, None]

    pair_surveys = []
    for filtername, filtername2 in zip(filter1s, filter2s):
        detailer_list = []
        bfs = []
        bfs.append(bf.M5DiffBasisFunction(filtername=filtername, nside=nside))
        if filtername2 is not None:
            bfs.append(bf.M5DiffBasisFunction(filtername=filtername2, nside=nside))
        bfs.append(
            bf.TargetMapBasisFunction(
                filtername=filtername,
                target_map=target_map[filtername],
                out_of_bounds_val=np.nan,
                nside=nside,
                norm_factor=norm_factor,
            )
        )
        if filtername2 is not None:
            bfs.append(
                bf.TargetMapBasisFunction(
                    filtername=filtername2,
                    target_map=target_map[filtername2],
                    out_of_bounds_val=np.nan,
                    nside=nside,
                    norm_factor=norm_factor,
                )
            )
        bfs.append(bf.SlewtimeBasisFunction(filtername=filtername, nside=nside))
        bfs.append(bf.StrictFilterBasisFunction(filtername=filtername))
        # Masks, give these 0 weight
        bfs.append(bf.AvoidDirectWind(nside=nside))
        bfs.append(bf.AltAzShadowMaskBasisFunction(nside=nside, shadow_minutes=60.0, max_alt=76.0))
        bfs.append(bf.MoonAvoidanceBasisFunction(nside=nside, moon_distance=30.0))
        bfs.append(bf.CloudedOutBasisFunction())
        # feasibility basis fucntions. Also give zero weight.
        filternames = [fn for fn in [filtername, filtername2] if fn is not None]
        bfs.append(bf.FilterLoadedBasisFunction(filternames=filternames))
        bfs.append(bf.TimeToTwilightBasisFunction(time_needed=22.0))
        bfs.append(bf.NotTwilightBasisFunction())
        bfs.append(bf.PlanetMaskBasisFunction(nside=nside))

        weights = np.array([3.0, 3.0, 0.3, 0.3, 3.0, 3.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        if filtername2 is None:
            # Need to scale weights up so filter balancing works properly.
            weights = np.array([6.0, 0.6, 3.0, 3.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        if filtername2 is None:
            survey_name = "blob, %s" % filtername
        else:
            survey_name = "blob, %s%s" % (filtername, filtername2)
        if filtername2 is not None:
            detailer_list.append(detailers.TakeAsPairsDetailer(filtername=filtername2))

        detailer_list.append(detailers.FlushByDetailer())
        pair_surveys.append(
            BlobSurvey(
                bfs,
                weights,
                filtername1=filtername,
                filtername2=filtername2,
                survey_note=survey_name,
                ignore_obs="DD",
                detailers=detailer_list,
                nside=nside,
            )
        )
    return pair_surveys


class TestExample(unittest.TestCase):
    @unittest.skipUnless(os.path.isfile(SAMPLE_BIG_DATA_FILE), "Test data not available.")
    def test_example(self):
        """Try out the example scheduler."""
        mjd_start = utils.survey_start_mjd()
        nside = 32
        survey_length = 4.0  # days
        scheduler = example_scheduler(nside=nside, mjd_start=mjd_start)
        observatory = ModelObservatory(nside=nside, mjd_start=mjd_start)
        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, survey_length=survey_length, filename=None
        )
        # Check that greedy observed some
        assert "greedy" in observations["note"]
        # check some long pairs got observed
        assert np.any(["pair_33" in obs for obs in observations["note"]])
        # Make sure lots of observations executed
        assert observations.size > 1000
        # Make sure nothing tried to look through the earth
        assert np.min(observations["alt"]) > 0
        # Make sure a DDF executed
        assert np.any(["DD" in note for note in observations["note"]])


class TestFeatures(unittest.TestCase):
    @unittest.skipUnless(os.path.isfile(SAMPLE_BIG_DATA_FILE), "Test data not available.")
    def test_greedy(self):
        """
        Set up a greedy survey and run for a few days.
        A crude way to touch lots of code.
        """
        mjd_start = utils.survey_start_mjd()
        nside = 32
        survey_length = 2.0  # days

        surveys = gen_greedy_surveys(nside)
        # Depreating Pairs_survey_scripted
        # surveys.append(Pairs_survey_scripted(None, ignore_obs='DD'))

        # Set up the DD
        dd_surveys = ddf_surveys(nside=nside)
        surveys.extend(dd_surveys)

        scheduler = CoreScheduler(surveys, nside=nside)
        observatory = ModelObservatory(nside=nside, mjd_start=mjd_start)
        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, survey_length=survey_length, filename=None
        )

        # Check that greedy observed some
        assert "greedy" in observations["note"]
        # Make sure lots of observations executed
        assert observations.size > 1000
        # Make sure nothing tried to look through the earth
        assert np.min(observations["alt"]) > 0

    @unittest.skipUnless(os.path.isfile(SAMPLE_BIG_DATA_FILE), "Test data not available.")
    def test_blobs(self):
        """
        Set up a blob selection survey
        """
        mjd_start = utils.survey_start_mjd()
        nside = 32
        survey_length = 2.0  # days

        surveys = []
        # Set up the DD
        dd_surveys = ddf_surveys(nside=nside)
        surveys.append(dd_surveys)

        surveys.append(gen_blob_surveys(nside))
        surveys.append(gen_greedy_surveys(nside))

        scheduler = CoreScheduler(surveys, nside=nside)
        observatory = ModelObservatory(nside=nside, mjd_start=mjd_start)
        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, survey_length=survey_length, filename=None
        )

        # Make sure some blobs executed
        assert "blob, gg, b" in observations["note"]
        assert "blob, gg, a" in observations["note"]
        # Make sure some greedy executed
        assert "greedy" in observations["note"]
        # Make sure lots of observations executed
        assert observations.size > 1000
        # Make sure nothing tried to look through the earth
        assert np.min(observations["alt"]) > 0

    @unittest.skipUnless(os.path.isfile(SAMPLE_BIG_DATA_FILE), "Test data not available.")
    def test_wind(self):
        """
        Test that a wind mask prevent things from being executed in the wrong spot
        """
        mjd_start = utils.survey_start_mjd()
        nside = 32
        survey_length = 4.0  # days

        surveys = []
        # Set up the DD
        dd_surveys = ddf_surveys(nside=nside)
        surveys.append(dd_surveys)

        surveys.append(gen_blob_surveys(nside))
        surveys.append(gen_greedy_surveys(nside))

        scheduler = CoreScheduler(surveys, nside=nside)
        observatory = ModelObservatoryWindy(nside=nside, mjd_start=mjd_start)
        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, survey_length=survey_length, filename=None
        )

        # Make sure some blobs executed
        assert "blob, gg, b" in observations["note"]
        assert "blob, gg, a" in observations["note"]
        # Make sure some greedy executed
        assert "greedy" in observations["note"]
        # Make sure lots of observations executed
        assert observations.size > 1000
        # Make sure nothing tried to look through the earth
        assert np.min(observations["alt"]) > 0

        # Make sure nothing executed in the strong wind
        assert np.min(np.degrees(observations["az"])) > 30.0
        assert np.max(np.degrees(observations["az"])) < (360.0 - 30.0)

    @unittest.skipUnless(os.path.isfile(SAMPLE_BIG_DATA_FILE), "Test data not available.")
    def test_nside(self):
        """
        test running at higher nside
        """
        mjd_start = utils.survey_start_mjd()
        nside = 64
        survey_length = 2.0  # days

        surveys = []
        # Set up the DD
        dd_surveys = ddf_surveys(nside=nside)
        surveys.append(dd_surveys)

        surveys.append(gen_blob_surveys(nside))
        surveys.append(gen_greedy_surveys(nside))

        scheduler = CoreScheduler(surveys, nside=nside)
        observatory = ModelObservatory(nside=nside, mjd_start=mjd_start)
        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, survey_length=survey_length, filename=None
        )

        # Make sure some blobs executed
        assert "blob, gg, b" in observations["note"]
        assert "blob, gg, a" in observations["note"]
        # Make sure some greedy executed
        assert "greedy" in observations["note"]
        # Make sure lots of observations executed
        assert observations.size > 1000
        # Make sure nothing tried to look through the earth
        assert np.min(observations["alt"]) > 0


if __name__ == "__main__":
    unittest.main()



__all__ = ("FieldSurvey",)

import numpy as np

from rubin_scheduler.utils import ra_dec2_hpid

from .dd_surveys import DeepDrillingSurvey


class FieldSurvey(DeepDrillingSurvey):
    def __init__(
        self,
        basis_functions,
        RA,
        dec,
        sequence="rgizy",
        nvis=[20, 10, 20, 26, 20],
        exptime=30.0,
        u_exptime=30.0,
        nexp=2,
        ignore_obs=None,
        survey_name="DD",
        reward_value=None,
        readtime=2.0,
        filter_change_time=120.0,
        nside=None,
        flush_pad=30.0,
        seed=42,
        detailers=None,
    ):
        super().__init__(
            basis_functions=basis_functions,
            RA=RA,
            dec=dec,
            sequence=sequence,
            nvis=nvis,
            exptime=exptime,
            u_exptime=u_exptime,
            nexp=nexp,
            ignore_obs=ignore_obs,
            survey_name=survey_name,
            reward_value=reward_value,
            readtime=readtime,
            filter_change_time=filter_change_time,
            nside=nside,
            flush_pad=flush_pad,
            seed=seed,
            detailers=detailers,
        )
        self.basis_weights = np.ones(len(basis_functions)) / len(basis_functions)
        self.ra_deg, self.dec_deg = RA, dec

    def calc_reward_function(self, conditions):
        self.reward_checked = True
        indx = ra_dec2_hpid(self.nside, self.ra_deg, self.dec_deg)
        if self._check_feasibility(conditions):
            self.reward = 0
            for bf, weight in zip(self.basis_functions, self.basis_weights):
                basis_value = bf(conditions, indx=indx)
                self.reward += basis_value * weight

            if not np.isscalar(self.reward):
                self.reward = np.sum(self.reward[indx])

                if np.any(np.isinf(self.reward)):
                    self.reward = np.inf
        else:
            # If not feasable, negative infinity reward
            self.reward = -np.inf
            return self.reward

        return self.reward



__all__ = ("Almanac",)

import datetime
import os

import numpy as np
from scipy.interpolate import interp1d

from rubin_scheduler.data import get_data_dir


class Almanac:
    """Class to load and return pre-computed information about the LSST site."""

    def __init__(self, mjd_start=None, kind="quadratic"):
        # Load up the sunrise/sunset times
        data_dir = os.path.join(get_data_dir(), "site_models")

        temp = np.load(os.path.join(data_dir, "sunsets.npz"))
        self.sunsets = temp["almanac"].copy()
        temp.close()

        if mjd_start is not None:
            loc = np.searchsorted(self.sunsets["sunset"], mjd_start)
            # Set the start MJD to be night 1.
            self.sunsets["night"] -= self.sunsets["night"][loc - 1]

        # Load the sun and moon positions
        temp = np.load(os.path.join(data_dir, "sun_moon.npz"))
        self.sun_moon = temp["sun_moon_info"].copy()
        temp.close()

        self.interpolators = {
            "sun_alt": interp1d(self.sun_moon["mjd"], self.sun_moon["sun_alt"], kind=kind),
            "sun_az_x": interp1d(self.sun_moon["mjd"], np.cos(self.sun_moon["sun_az"]), kind=kind),
            "sun_az_y": interp1d(self.sun_moon["mjd"], np.sin(self.sun_moon["sun_az"]), kind=kind),
            "sun_RA_x": interp1d(self.sun_moon["mjd"], np.cos(self.sun_moon["sun_RA"]), kind=kind),
            "sun_RA_y": interp1d(self.sun_moon["mjd"], np.sin(self.sun_moon["sun_RA"]), kind=kind),
            "sun_dec": interp1d(self.sun_moon["mjd"], self.sun_moon["sun_dec"], kind=kind),
            "moon_alt": interp1d(self.sun_moon["mjd"], self.sun_moon["moon_alt"], kind=kind),
            "moon_az_x": interp1d(self.sun_moon["mjd"], np.cos(self.sun_moon["moon_az"]), kind=kind),
            "moon_az_y": interp1d(self.sun_moon["mjd"], np.sin(self.sun_moon["moon_az"]), kind=kind),
            "moon_RA_x": interp1d(self.sun_moon["mjd"], np.cos(self.sun_moon["moon_RA"]), kind=kind),
            "moon_RA_y": interp1d(self.sun_moon["mjd"], np.sin(self.sun_moon["moon_RA"]), kind=kind),
            "moon_dec": interp1d(self.sun_moon["mjd"], self.sun_moon["moon_dec"], kind=kind),
            "moon_phase": interp1d(self.sun_moon["mjd"], self.sun_moon["moon_phase"], kind=kind),
        }

        temp = np.load(os.path.join(data_dir, "planet_locations.npz"))
        self.planet_loc = temp["planet_loc"].copy()
        temp.close()

        self.planet_names = ["venus", "mars", "jupiter", "saturn"]
        self.planet_interpolators = {}
        for pn in self.planet_names:
            self.planet_interpolators[pn + "_RA_x"] = interp1d(
                self.planet_loc["mjd"], np.cos(self.planet_loc[pn + "_RA"]), kind=kind
            )
            self.planet_interpolators[pn + "_RA_y"] = interp1d(
                self.planet_loc["mjd"], np.sin(self.planet_loc[pn + "_RA"]), kind=kind
            )
            self.planet_interpolators[pn + "_dec"] = interp1d(
                self.planet_loc["mjd"], self.planet_loc[pn + "_dec"], kind=kind
            )

    def get_planet_positions(self, mjd):
        result = {}
        for pn in self.planet_names:
            result[pn + "_dec"] = self.planet_interpolators[pn + "_dec"](mjd)

            temp_result = np.array(
                [
                    np.arctan2(
                        self.planet_interpolators[pn + "_RA_y"](mjd),
                        self.planet_interpolators[pn + "_RA_x"](mjd),
                    )
                ]
            ).ravel()
            negative_angles = np.where(temp_result < 0.0)[0]
            temp_result[negative_angles] = 2.0 * np.pi + temp_result[negative_angles]
            result[pn + "_RA"] = temp_result
        return result

    def get_sunset_info(self, mjd=None, evening_date=None, longitude=None):
        """Returns a numpy array with mjds for various events
        (sunset, moonrise, sun at -12 degrees alt, etc.).

        Parameters
        ----------
        mjd : `float`
            A UTC MJD that occurs during the desired night.
            Defaults to None.
        evening_date : `str` or `datetime.date`
            The local date of the evening of the night whose index is desired,
            in ISO8601 format (YYYY-MM-DD). Defaults to None.
        longitude : `float` or  `astropy.coordinates.angles.core.Angle`
            If a float, then the value is interpreted as being in radians.
            Defaults to None.

        Returns
        -------
        sunset_info : `numpy.void`
            A numpy object with dtype([
                ('night', '<i8'),
                ('sunset', '<f8'),
                ('sun_n12_setting', '<f8'),
                ('sun_n18_setting', '<f8'),
                ('sun_n18_rising', '<f8'),
                ('sun_n12_rising', '<f8'),
                ('sunrise', '<f8'),
                ('moonrise', '<f8'),
                ('moonset', '<f8')
            ])
        """
        if mjd is not None and evening_date is not None:
            raise ValueError("At most one of mjd and evening_date can be set")

        # Default to now
        if mjd is None and evening_date is None:
            mjd = 40587 + datetime.datetime.now().timestamp() / (24 * 60 * 60)

        if mjd is not None:
            indx = np.searchsorted(self.sunsets["sunset"], mjd, side="right") - 1
        elif evening_date is not None:
            if longitude is None:
                raise ValueError("If evening_date is set, longitude is needed as well")
            indx = self.index_for_local_evening(evening_date, longitude)

        return self.sunsets[indx]

    def mjd_indx(self, mjd):
        indx = np.searchsorted(self.sunsets["sunset"], mjd, side="right") - 1
        return indx

    def index_for_local_evening(self, evening_date, longitude):
        """The index of the night with sunset at a given local date.

        Parameters
        ----------
        evening_date : `str` or `datetime.date`
            The local date of the evening of the night whose index is desired,
            in ISO8601 format (YYYY-MM-DD).
        longitude : `float` or  `astropy.coordinates.angles.core.Angle`
            If a float, then the value is interpreted as being in radians.

        Returns
        -------
        night_index : `int`
            The index of the requested night.
        """
        try:
            longitude = longitude.radian
        except AttributeError:
            pass

        if isinstance(evening_date, str):
            evening_date = datetime.date.fromisoformat(evening_date)

        evening_datetime = datetime.datetime(evening_date.year, evening_date.month, evening_date.day)
        evening_mjd = np.floor(evening_datetime.timestamp() / (24 * 60 * 60) + 40587)

        # Depending on the time of year, the UTC date rollover might not
        # always be on the same side of local sunset. Shift by the longitude
        # to make sure the rollover is always near midnight, far from sunset.
        matching_nights = np.argwhere(
            np.floor(self.sunsets["sunset"] + longitude / (2 * np.pi)) == evening_mjd
        )
        if len(matching_nights) < 1:
            raise ValueError(f"Requested night {evening_date} outside of almanac date range")

        night_index = matching_nights.item()

        return night_index

    def get_sun_moon_positions(self, mjd):
        """
        All angles in Radians. moonPhase between 0 and 100.
        """
        simple_calls = ["sun_alt", "sun_dec", "moon_alt", "moon_dec", "moon_phase"]
        result = {}
        for key in simple_calls:
            result[key] = self.interpolators[key](mjd)

        longitude_calls = ["sun_az", "moon_az", "sun_RA", "moon_RA"]
        for key in longitude_calls:
            # Need to wrap in case sent a scalar
            temp_result = np.array(
                [
                    np.arctan2(
                        self.interpolators[key + "_y"](mjd),
                        self.interpolators[key + "_x"](mjd),
                    )
                ]
            ).ravel()
            negative_angles = np.where(temp_result < 0.0)[0]
            temp_result[negative_angles] = 2.0 * np.pi + temp_result[negative_angles]
            result[key] = temp_result

        return result



__all__ = ("Conditions",)

import functools
import warnings
from io import StringIO

import healpy as hp
import numpy as np
import pandas as pd

from rubin_scheduler.scheduler.utils import (
    match_hp_resolution,
    season_calc,
    set_default_nside,
    smallest_signed_angle,
)
from rubin_scheduler.utils import (
    Site,
    _angular_separation,
    _approx_altaz2pa,
    _approx_ra_dec2_alt_az,
    _hpid2_ra_dec,
    calc_lmst,
    m5_flat_sed,
)


class Conditions:
    """
    Class to hold telemetry information

    If the incoming value is a healpix map, we use a setter to ensure the
    resolution matches.

    Unless otherwise noted, all values are assumed to be valid at the time
    given by self.mjd
    """

    global_maps = set(["ra", "dec", "slewtime", "airmass", "zeros_map", "nan_map"])
    by_band_maps = set(["skybrightness", "fwhm_eff", "m5_depth"])

    def __init__(
        self,
        nside=None,
        site="LSST",
        exptime=30.0,
        mjd_start=59853.5,
        season_offset=None,
        sun_ra_start=None,
        mjd=None,
    ):
        """
        Parameters
        ----------
        nside : int
            The healpixel nside to set the resolution of attributes.
        site : str ('LSST')
            A site name used to create a sims.utils.Site object. For looking up
            observatory paramteres like latitude and longitude.
        expTime : float (30)
            The exposure time to assume when computing the 5-sigma limiting depth
        mjd_start : float (59853.5)
            The starting MJD of the survey.
        season_offset : np.array
            A HEALpix array that specifies the day offset when computing the season for each HEALpix.
        sun_ra_start : float (None)
            The RA of the sun at the start of the survey (radians)
        mjd : float
            The current MJD.

        Attributes (Set on init)
        -----------
        nside : int
            Healpix resolution. All maps are set to this reslution.
        site : rubin_scheduler.Site object ('LSST')
            Contains static site-specific data (lat, lon, altitude, etc). Defaults to 'LSST'.
        ra : np.array
            A healpix array with the RA of each healpixel center (radians). Automatically
            generated.
        dec : np.array
            A healpix array with the Dec of each healpixel center (radians). Automatically generated.

        Attributes (to be set by user/telemetry stream/scheduler)
        -------------------------------------------
        mjd : float
            Modified Julian Date (days).
        bulk_cloud : float
            The fraction of sky covered by clouds. (In the future might update to transparency map)
        cloud_map : np.array
            XXX--to be done. HEALpix array with cloudy pixels set to NaN.
        slewtime : np.array
            Healpix showing the slewtime to each healpixel center (seconds)
        current_filter : str
            The name of the current filter. (expect one of u, g, r, i, z, y).
        mounted_filters : list of str
            The filters that are currently mounted and thu available (expect 5 of u, g, r, i, z, y)
        night : int
            The current night number (days). Probably starts at 1.
        skybrightness : dict of np.array
            Dictionary keyed by filtername. Values are healpix arrays with the sky brightness at each
            healpix center (mag/acsec^2)
        fwhm_eff : dict of np.array
            Dictionary keyed by filtername. Values are the effective seeing FWHM at each healpix
            center (arcseconds)
        moon_alt : float
            The altitude of the Moon (radians)
        moon_az : float
            The Azimuth of the moon (radians)
        moon_ra : float
            RA of the moon (radians)
        moon_dec : float
            Declination of the moon (radians)
        moon_phase : float
            The Phase of the moon. (percent, 0=new moon, 100=full moon)
        sun_alt : float
            The altitude of the sun (radians).
        sun_az : float
            The Azimuth of the sun (radians).
        sun_ra : float
            The RA of the sun (radians).
        sun_dec : float
            The Dec of the sun (radians).
        tel_ra : float
            The current telescope RA pointing (radians).
        tel_dec : float
            The current telescope Declination (radians).
        tel_alt : float
            The current telescope altitude (radians).
        tel_az : float
            The current telescope azimuth (radians).
        cumulative_azimuth_rad : float
            The cummulative telescope azimuth (radians). For tracking cable wrap
        cloud_map : np.array
            A healpix map with the cloud coverage. XXX-expand, is this bool map? Transparency map?
        airmass : np.array
            A healpix map with the airmass value of each healpixel. (unitless)
        wind_speed : float
            Wind speed (m/s).
        wind_direction : float
            Direction from which the wind originates. A direction of 0.0 degrees
            means the wind originates from the north and 90.0 degrees from the
            east (radians).
        sunset : float
            The MJD of sunset that starts the current night. Note MJDs of sunset, moonset, twilight times, etc
            are from interpolations. This means the sun may actually be slightly above/below the horizon
            at the given sunset time.
        sun_n12_setting : float
            The MJD of when the sun is at -12 degees altitude and setting during the
            current night. From interpolation.
        sun_n18_setting : float
            The MJD when the sun is at -18 degrees altitude and setting during the current night.
            From interpolation.
        sun_n18_rising : float
            The MJD when the sun is at -18 degrees altitude and rising during the current night.
            From interpolation.
        sun_n12_rising : float
            The MJD when the sun is at -12 degrees altitude and rising during the current night.
            From interpolation.
        sunrise : float
            The MJD of sunrise during the current night. From interpolation
        mjd_start : float
            The starting MJD of the survey.
        moonrise : float
            The MJD of moonrise during the current night. From interpolation.
        moonset : float
            The MJD of moonset during the current night. From interpolation.
        targets_of_opportunity : list of rubin_scheduler.scheduler.targetoO object(s)
            targetoO objects.
        planet_positions : dict
            Dictionary of planet name and coordinate e.g., 'venus_RA', 'mars_dec'
        scheduled_observations : np.array
            A list of MJD times when there are scheduled observations. Defaults to empty array.
        tel_az_limits : list of float pairs
            A list of lists giving valid azimuth ranges. e.g., [0, 2*np.pi] would
            mean all azimuth values are valid, while [[0, np.pi/2], [3*np.pi/2, 2*np.pi]]
            would mean anywhere in the south is invalid.  Radians.
        tel_alt_limits : list of float pairs
            A list of lists giving valid altitude ranges. Radians.

        Attributes (calculated on demand and cached)
        ------------------------------------------
        alt : np.array
            Altitude of each healpixel (radians). Recaclulated if mjd is updated. Uses fast
            approximate equation for converting RA,Dec to alt,az.
        az : np.array
            Azimuth of each healpixel (radians). Recaclulated if mjd is updated. Uses fast
            approximate equation for converting RA,Dec to alt,az.
        pa : np.array
            The parallactic angle of each healpixel (radians). Recaclulated if mjd is updated.
            Based on the fast approximate alt,az values.
        lmst : float
            The local mean sidearal time (hours). Updates is mjd is changed.
        m5_depth : dict of np.array
            the 5-sigma limiting depth healpix maps, keyed by filtername (mags). Will be recalculated
            if the skybrightness, seeing, or airmass are updated.
        HA : np.array
            Healpix map of the hour angle of each healpixel (radians). Runs from 0 to 2pi.
        az_to_sun : np.array
            Healpix map of the azimuthal distance to the sun for each healpixel (radians)
        az_to_anitsun : np.array
            Healpix map of the azimuthal distance to the anit-sun for each healpixel (radians)
        solar_elongation : np.array
            Healpix map of the solar elongation (angular distance to the sun) for each healpixel (radians)

        Attributes (set by the scheduler)
        -------------------------------
        queue : list of observation objects
            The current queue of observations core_scheduler is waiting to execute.

        """
        if nside is None:
            nside = set_default_nside()
        self.nside = nside
        self.site = Site(site)
        self.exptime = exptime
        self.mjd_start = mjd_start
        hpids = np.arange(hp.nside2npix(nside))
        self.season_offset = season_offset
        self.sun_ra_start = sun_ra_start
        # Generate an empty map so we can copy when we need a new map
        self.zeros_map = np.zeros(hp.nside2npix(nside), dtype=float)
        self.nan_map = np.zeros(hp.nside2npix(nside), dtype=float)
        self.nan_map.fill(np.nan)
        # The RA, Dec grid we are using
        self.ra, self.dec = _hpid2_ra_dec(nside, hpids)

        self._init_attributes()
        self.mjd = mjd

    def _init_attributes(self):
        """Initialize all the attributes"""

        # Modified Julian Date (day)
        self._mjd = None
        # Altitude and azimuth. Dict with degrees and radians
        self._alt = None
        self._az = None
        self._pa = None
        # The cloud level. Fraction, but could upgrade to transparency map
        self.clouds = None
        self._slewtime = None
        self.current_filter = None
        self.mounted_filters = None
        self.night = None
        self._lmst = None
        # Should be a dict with filtername keys
        self._skybrightness = {}
        self._fwhm_eff = {}
        self._m5_depth = None
        self._airmass = None

        self.wind_speed = None
        self.wind_direction = None

        # Upcomming scheduled observations
        self.scheduled_observations = np.array([], dtype=float)

        # Attribute to hold the current observing queue
        self.queue = None

        # Moon
        self.moon_alt = None
        self.moon_az = None
        self.moon_ra = None
        self.moon_dec = None
        self.moon_phase = None

        # Sun
        self.sun_alt = None
        self.sun_az = None
        self.sun_ra = None
        self.sun_dec = None

        # Almanac information
        self.sunset = None
        self.sun_n12_setting = None
        self.sun_n18_setting = None
        self.sun_n18_rising = None
        self.sun_n12_rising = None
        self.sunrise = None
        self.moonrise = None
        self.moonset = None

        self.planet_positions = None

        # Current telescope pointing
        self.tel_ra = None
        self.tel_dec = None
        self.tel_alt = None
        self.tel_az = None
        self.cumulative_azimuth_rad = None

        # Telescope limits
        self.tel_az_limits = None
        self.tel_alt_limits = None

        # Full sky cloud map
        self._cloud_map = None
        self._HA = None

        # XXX--document
        self.bulk_cloud = None

        self.rot_tel_pos = None

        self.targets_of_opportunity = None

        self._season = None

        self.season_modulo = None
        self.season_max_season = None
        self.season_length = 365.25
        self.season_floor = True

        # Potential attributes that get computed
        self._solar_elongation = None
        self._az_to_sun = None
        self._az_to_antisun = None

    @property
    def lmst(self):
        if self._lmst is None:
            self._lmst = calc_lmst(self.mjd, self.site.longitude_rad)

        return self._lmst

    @lmst.setter
    def lmst(self, value):
        self._lmst = value
        self._HA = None

    @property
    def HA(self):
        if self._HA is None:
            self.calc_ha()
        return self._HA

    def calc_ha(self):
        self._HA = np.radians(self.lmst * 360.0 / 24.0) - self.ra
        self._HA[np.where(self._HA < 0)] += 2.0 * np.pi

    @property
    def cloud_map(self):
        return self._cloud_map

    @cloud_map.setter
    def cloud_map(self, value):
        self._cloud_map = match_hp_resolution(value, nside_out=self.nside)

    @property
    def slewtime(self):
        return self._slewtime

    @slewtime.setter
    def slewtime(self, value):
        # Using 0 for start of night
        if np.size(value) == 1:
            self._slewtime = value
        else:
            self._slewtime = match_hp_resolution(value, nside_out=self.nside)

    @property
    def airmass(self):
        return self._airmass

    @airmass.setter
    def airmass(self, value):
        self._airmass = match_hp_resolution(value, nside_out=self.nside)
        self._m5_depth = None

    @property
    def pa(self):
        if self._pa is None:
            self.calc_pa()
        return self._pa

    def calc_pa(self):
        self._pa = _approx_altaz2pa(self.alt, self.az, self.site.latitude_rad)

    @property
    def alt(self):
        if self._alt is None:
            self.calc_alt_az()
        return self._alt

    @property
    def az(self):
        if self._az is None:
            self.calc_alt_az()
        return self._az

    def calc_alt_az(self):
        self._alt, self._az = _approx_ra_dec2_alt_az(
            self.ra,
            self.dec,
            self.site.latitude_rad,
            self.site.longitude_rad,
            self._mjd,
        )

    @functools.lru_cache(maxsize=10)
    def future_alt_az(self, mjd):
        """Compute the altitude and azimuth for a future time.

        Returns
        -------
        altitude : `np.array`
            The altutude of each healpix at MJD (radians)
        azimuth : : `np.array`
            The azimuth of each healpix at MJD (radians)
        """
        alt, az = _approx_ra_dec2_alt_az(
            self.ra,
            self.dec,
            self.site.latitude_rad,
            self.site.longitude_rad,
            mjd,
        )
        return alt, az

    @property
    def mjd(self):
        return self._mjd

    @mjd.setter
    def mjd(self, value):
        # If MJD is changed, everything else is no longer valid, so re-init
        if self.mjd is not None:
            warnings.warn("Changing MJD and resetting all attributes.")
        self._init_attributes()
        self._mjd = value

    @property
    def skybrightness(self):
        return self._skybrightness

    @skybrightness.setter
    def skybrightness(self, indict):
        for key in indict:
            self._skybrightness[key] = match_hp_resolution(indict[key], nside_out=self.nside)
        # If sky brightness changes, need to recalc M5 depth.
        self._m5_depth = None

    @property
    def fwhm_eff(self):
        return self._fwhm_eff

    @fwhm_eff.setter
    def fwhm_eff(self, indict):
        for key in indict:
            self._fwhm_eff[key] = match_hp_resolution(indict[key], nside_out=self.nside)
        self._m5_depth = None

    @property
    def m5_depth(self):
        if self._m5_depth is None:
            self.calc_m5_depth()
        return self._m5_depth

    def calc_m5_depth(self):
        self._m5_depth = {}
        for filtername in self._skybrightness:
            good = ~np.isnan(self._skybrightness[filtername])
            self._m5_depth[filtername] = self.nan_map.copy()
            self._m5_depth[filtername][good] = m5_flat_sed(
                filtername,
                self._skybrightness[filtername][good],
                self._fwhm_eff[filtername][good],
                self.exptime,
                self._airmass[good],
            )

    def calc_solar_elongation(self):
        self._solar_elongation = _angular_separation(self.ra, self.dec, self.sun_ra, self.sun_dec)

    @property
    def solar_elongation(self):
        if self._solar_elongation is None:
            self.calc_solar_elongation()
        return self._solar_elongation

    def calc_az_to_sun(self):
        self._az_to_sun = smallest_signed_angle(self.ra, self.sun_ra)

    def calc_az_to_antisun(self):
        self._az_to_antisun = smallest_signed_angle(self.ra + np.pi, self.sun_ra)

    @property
    def az_to_sun(self):
        if self._az_to_sun is None:
            self.calc_az_to_sun()
        return self._az_to_sun

    @property
    def az_to_antisun(self):
        if self._az_to_antisun is None:
            self.calc_az_to_antisun()
        return self._az_to_antisun

    # XXX, there's probably an elegant decorator that could do this caching automatically
    def season(self, modulo=None, max_season=None, season_length=365.25, floor=True):
        if self.season_offset is not None:
            kwargs_match = (
                (modulo == self.season_modulo)
                & (max_season == self.season_max_season)
                & (season_length == self.season_length)
                & (floor == self.season_floor)
            )
            if ~kwargs_match:
                self.season_modulo = modulo
                self.season_max_season = max_season
                self.season_length = season_length
                self.season_floor = floor
            if (self._season is None) | (~kwargs_match):
                self._season = season_calc(
                    self.night,
                    offset=self.season_offset,
                    modulo=modulo,
                    max_season=max_season,
                    season_length=season_length,
                    floor=floor,
                )
        else:
            self._season = None

        return self._season

    def __repr__(self):
        return f"<{self.__class__.__name__} mjd_start='{self.mjd_start}' at {hex(id(self))}>"

    def __str__(self):
        # If dependencies of to_markdown are not installed, fall back on repr
        try:
            pd.DataFrame().to_markdown()
        except ImportError:
            return repr(self)

        if self.mjd is None:
            # Many elements of the pretty str representation rely on
            # the time beging set. If it is not, just return the ugly form.
            return repr(self)

        output = StringIO()
        print(f"{self.__class__.__qualname__} at {hex(id(self))}", file=output)
        print("============================", file=output)
        print("nside: ", self.nside, "  ", file=output)
        print("site: ", self.site.name, "  ", file=output)
        print("exptime: ", self.exptime, "  ", file=output)
        print("lmst: ", self.lmst, "  ", file=output)
        print("season_offset: ", self.season_offset, "  ", file=output)
        print("sun_RA_start: ", self.sun_ra_start, "  ", file=output)
        print("clouds: ", self.clouds, "  ", file=output)
        print("current_filter: ", self.current_filter, "  ", file=output)
        print("mounted_filters: ", self.mounted_filters, "  ", file=output)
        print("night: ", self.night, "  ", file=output)
        print("wind_speed: ", self.wind_speed, "  ", file=output)
        print("wind_direction: ", self.wind_direction, "  ", file=output)
        print(
            "len(scheduled_observations): ",
            len(self.scheduled_observations),
            "  ",
            file=output,
        )
        print(
            "len(queue): ",
            None if self.queue is None else len(self.queue),
            "  ",
            file=output,
        )
        print("moonPhase: ", self.moon_phase, "  ", file=output)
        print("bulk_cloud: ", self.bulk_cloud, "  ", file=output)
        print("targets_of_opportunity: ", self.targets_of_opportunity, "  ", file=output)
        print("season_modulo: ", self.season_modulo, "  ", file=output)
        print("season_max_season: ", self.season_max_season, "  ", file=output)
        print("season_length: ", self.season_length, "  ", file=output)
        print("season_floor: ", self.season_floor, "  ", file=output)
        print("cumulative_azimuth_rad: ", self.cumulative_azimuth_rad, "  ", file=output)

        positions = [
            {
                "name": "sun",
                "alt": self.sun_alt,
                "az": self.sun_az,
                "RA": self.sun_ra,
                "decl": self.sun_dec,
            }
        ]
        positions.append(
            {
                "name": "moon",
                "alt": self.moon_alt,
                "az": self.moon_az,
                "RA": self.moon_ra,
                "decl": self.moon_dec,
            }
        )
        for planet_name in ("venus", "mars", "jupiter", "saturn"):
            positions.append(
                {
                    "name": planet_name,
                    "RA": float(self.planet_positions[planet_name + "_RA"]),
                    "decl": float(self.planet_positions[planet_name + "_dec"]),
                }
            )
        positions.append(
            {
                "name": "telescope",
                "alt": self.tel_alt,
                "az": self.tel_az,
                "RA": self.tel_ra,
                "decl": self.tel_dec,
                "rot": self.rot_tel_pos,
            }
        )
        positions = pd.DataFrame(positions).set_index("name")
        print(file=output)
        print("Positions (radians)", file=output)
        print("-------------------", file=output)
        print(positions.to_markdown(), file=output)

        positions_deg = np.degrees(positions)
        print(file=output)
        print("Positions (degrees)", file=output)
        print("-------------------", file=output)
        print(positions_deg.to_markdown(), file=output)

        events = (
            "mjd_start",
            "mjd",
            "sunset",
            "sun_n12_setting",
            "sun_n18_setting",
            "sun_n18_rising",
            "sun_n12_rising",
            "sunrise",
            "moonrise",
            "moonset",
            "sun_0_setting",
            "sun_0_rising",
        )
        event_rows = []
        for event in events:
            try:
                mjd = getattr(self, event)
                time = pd.to_datetime(float(mjd) + 2400000.5, unit="D", origin="julian")
                event_rows.append({"event": event, "MJD": mjd, "date": time})
            except AttributeError:
                pass

        event_df = pd.DataFrame(event_rows).set_index("event").sort_values(by="MJD")
        print("", file=output)
        print("Events", file=output)
        print("------", file=output)
        print(event_df.to_markdown(), file=output)

        map_stats = []
        for map_name in self.global_maps - set(["zeros_map", "nan_map"]):
            try:
                values = getattr(self, map_name)
                map_stats.append(
                    {
                        "map": map_name,
                        "nside": hp.npix2nside(len(values)),
                        "min": np.nanmin(values),
                        "max": np.nanmax(values),
                        "median": np.nanmedian(values),
                    }
                )
            except AttributeError:
                pass

        for base_map_name in self.by_band_maps:
            for band in "ugrizy":
                try:
                    values = getattr(self, base_map_name)[band]
                    map_name = f"{base_map_name}_{band}"
                    map_stats.append(
                        {
                            "map": map_name,
                            "nside": hp.npix2nside(len(values)),
                            "min": np.nanmin(values),
                            "max": np.nanmax(values),
                            "median": np.nanmedian(values),
                        }
                    )
                except AttributeError:
                    pass

        maps_df = pd.DataFrame(map_stats).set_index("map")
        print("", file=output)
        print("Maps", file=output)
        print("----", file=output)
        print(maps_df.to_markdown(), file=output)

        result = output.getvalue()
        return result

    def _repr_markdown_(self):
        return str(self)



__all__ = (
    "order_observations",
    "generate_dist_matrix",
    "route_length",
    "generate_hulls",
    "merge_hulls",
    "three_opt",
    "tsp_convex",
)

import itertools
from collections import deque

import numpy as np
import scipy.spatial as spatial

from .utils import IntRounded, gnomonic_project_toxy, mean_azimuth

# Solve Traveling Salesperson using convex hulls.
# re-write of https://github.com/jameskrysiak/ConvexSalesman/blob/master/convex_salesman.py
# This like a good explination too https://www.youtube.com/watch?v=syRSy1MFuho


def order_observations(lon, lat, scale=1e6, optimize=False):
    """Use TSP solver to put observations in an order that minimizes
    angular distance traveled

    Parameters
    ----------
    lon : `float`
        A longitude-like (RA, azimuth) angle (radians).
    lat : `float`
        A latitude-like (dec, altitude) angle (radians).
    scale : `float`
        A factor to scale and round projections to force same
        machine precision cross-platforms (1e6).
    optimize : `bool`
        If the TSP should run extra optimization steps, default False
    """

    # Let's find a good spot to project the points to a plane
    mid_dec = (np.max(lat) - np.min(lat)) / 2.0 + np.min(lat)
    mid_ra = mean_azimuth(lon)
    # Project the coordinates to a plane. Could consider scaling things to represent
    # time between points rather than angular distance.
    pointing_x, pointing_y = gnomonic_project_toxy(lon, lat, mid_ra, mid_dec)
    # Round off positions so that we ensure identical cross-platform performance

    pointing_x = np.round(pointing_x * scale).astype(int)
    pointing_y = np.round(pointing_y * scale).astype(int)
    # Now I have a bunch of x,y pointings. Drop into TSP solver to get an effiencent route
    towns = np.vstack((pointing_x, pointing_y)).T
    # Leaving optimize=False for speed. The optimization step doesn't usually improve much.
    better_order = tsp_convex(towns, optimize=optimize)
    return better_order


def generate_dist_matrix(towns):
    """Generate the matrix for the distance between town i and j

    Parameters
    ----------
    towns : np.array
        The x,y positions of the towns
    """

    x = towns[:, 0]
    y = towns[:, 1]
    # Broadcast to i,j
    x_dist = x - x[:, np.newaxis]
    y_dist = y - y[:, np.newaxis]
    distances = np.sqrt(x_dist**2 + y_dist**2)

    return distances


def route_length(town_indx, dist_matrix):
    """Find the length of a route

    Parameters
    ----------
    town_indx : array of int
        The indices of the towns.
    dist_matrix : np.array
        The matrix where the (i,j) elements are the distance
        between the ith and jth town
    """

    # This closes the path and return to the start
    town_i = town_indx
    town_j = np.roll(town_indx, -1)
    distances = dist_matrix[town_i, town_j]
    return np.sum(distances)


def generate_hulls(towns):
    """Given an array of x,y points, sort them into concentric hulls

    Parameters
    ----------
    towns : np.array (n,2)
        Array of town x,y positions

    Returns
    -------
    list of lists of the indices of the concentric hulls
    """

    # The indices we have to sort
    all_indices = np.arange(towns.shape[0])
    # array to note if a town has been used in a hull
    indices_used = np.zeros(towns.shape[0], dtype=bool)
    results = []

    # Continue until every point is inside a convex hull.
    while False in indices_used:
        # Try to find the convex hull of the remaining points.
        try:
            new_hull = spatial.ConvexHull(towns[all_indices[~indices_used]])
            new_indices = all_indices[~indices_used][new_hull.vertices]
            results.append(new_indices.tolist())
            indices_used[new_indices] = True

        # In a degenerate case (fewer than three points, points collinear)
        # Add all of the remaining points to the innermost convex hull.
        except spatial._qhull.QhullError:
            results.append(all_indices[~indices_used].tolist())
            indices_used[~indices_used] = True
            return results

    return results


def merge_hulls(indices_lists, dist_matrix):
    """Combine the hulls

    Parameters
    ----------
    indices_list : list of lists with ints
    dist_matrix : np.array
    """
    # start with the outer hull one. Use deque to rotate fast.
    collapsed_indices = deque(indices_lists[0])
    for ind_list in indices_lists[1:]:
        # insert each point indvidually
        for indx in ind_list:
            possible_results = []
            possible_lengths = []
            dindex = deque([indx])
            # In theory, I think this could loop over fewer points. Only need to check
            # points that can "see" the inner points?
            for i in range(len(collapsed_indices)):
                collapsed_indices.rotate(1)
                possible_results.append(collapsed_indices + dindex)
                possible_lengths.append(route_length(possible_results[-1], dist_matrix))
            best = np.min(np.where(possible_lengths == np.nanmin(possible_lengths)))
            collapsed_indices = possible_results[best]
    return list(collapsed_indices)


def three_opt(route, dist_matrix, cross_platform=True):
    """Iterates over all possible 3-optional transformations.

    Parameters
    ----------
    route : `list`
        The indices of the route
    dist_matrix : `np.array`
        Distance matrix for the towns
    cross_platform : `bool`
        Use utils.intRounded to make sure results will be repeatable
        cross-platform. Default True

    Returns
    -------
    min_route : list
        The new route
    min_length : float
        The length of the new route

    """
    # The combinations of three places that we can split each route.
    combinations = list(itertools.combinations(range(len(route)), 3))

    min_route = route
    min_length = route_length(min_route, dist_matrix)

    for cuts in combinations:
        # The three chunks that the route is broken into based on the cuts.
        c1 = route[cuts[0] : cuts[1]]
        c2 = route[cuts[1] : cuts[2]]
        c3 = route[cuts[2] :] + route[: cuts[0]]

        # Reversed chunks 2 and 3.
        rc2 = c2[::-1]
        rc3 = c3[::-1]

        # The unique permutations of all of those chunks.
        route_perms = [
            c1 + c2 + c3,
            c1 + c3 + c2,
            c1 + rc2 + c3,
            c1 + c3 + rc2,
            c1 + c2 + rc3,
            c1 + rc3 + c2,
            c1 + rc2 + rc3,
            c1 + rc3 + rc2,
        ]

        # Find the smallest of these permutations.
        for perm in route_perms:
            temp_length = route_length(perm, dist_matrix)
            if cross_platform:
                if IntRounded(temp_length) < IntRounded(min_length):
                    min_length = temp_length
                    min_route = perm
            else:
                if temp_length < min_length:
                    min_length = temp_length
                    min_route = perm

    return min_route, min_length


def tsp_convex(towns, optimize=False, niter=10):
    """Find a route through towns

    Parameters
    ----------
    towns : np.array (shape n,2)
        The points to find a path through
    optimize : bool (False)
        Optional to run the 3-optional transformation to optimize route
    niter : int (10)
        Max number of iterations to run on optimize loop.

    Returns
    -------
    indices that order towns.
    """
    hull_verts = generate_hulls(towns)
    dist_matrix = generate_dist_matrix(towns)
    route = merge_hulls(hull_verts, dist_matrix)

    if optimize:
        distance = route_length(route, dist_matrix)
        iter_count = 0
        optimized = False
        while not optimized:
            new_route, new_distance = three_opt(route, dist_matrix)
            if new_distance < distance:
                route = new_route
                distance = new_distance
                iter_count += 1
            else:
                optimized = True
            if iter_count == niter:
                return route
    return route



__all__ = ("m5_flat_sed", "m5_scale")

import numpy as np

from .sys_eng_vals import SysEngVals


def m5_scale(
    exp_time,
    nexp,
    airmass,
    fwhm_eff,
    musky,
    dark_sky_mag,
    cm,
    d_cm_infinity,
    k_atm,
    tau_cloud=0,
    base_exp_time=15,
):
    """Return m5 (scaled) value for all filters.

    Parameters
    ----------
    exp_time : float
        Exposure time (in seconds) for each exposure
    nexp : int
        Number of exposures
    airmass : float
        Airmass of the observation
    fwhm_eff : np.ndarray or pd.DataFrame
        FWHM (in arcseconds) per filter
    musky : np.ndarray or pd.DataFrame
        Sky background (in magnitudes/sq arcsecond) per filter of the observation
    dark_sky_mag : np.ndarray or pd.DataFrame
        Dark Sky, zenith magnitude/sq arcsecond - to scale musky. per filter
    cm : np.ndarray or pd.DataFrame
        cm value for the throughputs per filter
    d_cm_infinity : np.ndarray or pd.DataFrame
        d_cm_infinity values for the throughputs, per filter
    k_atm : np.ndarray or pd.DataFrame
        Atmospheric extinction values, per filter
    tau_cloud : float, optional
        Extinction due to clouds
    base_exp_time : float, optional
        The exposure time used to calculate cm / d_cm_infinity. Used to scale exp_time.
        This is the individual exposure exposure time.

    Returns
    -------
    np.ndarray or pd.DataFrame
        m5 values scaled for the visit conditions

    Note: The columns required as input for m5_scale can be calculated using
    the makeM5 function in lsst.syseng.throughputs.
    """
    # Calculate adjustment if readnoise is significant for exposure time
    # (see overview paper, equation 7)
    tscale = exp_time / base_exp_time * np.power(10.0, -0.4 * (musky - dark_sky_mag))
    d_cm = 0.0
    d_cm += d_cm_infinity
    d_cm -= 1.25 * np.log10(1 + (10 ** (0.8 * d_cm_infinity) - 1) / tscale)
    # Calculate m5 for 1 exp - constants here come from definition of cm/d_cm_infinity
    m5 = (
        cm
        + d_cm
        + 0.50 * (musky - 21.0)
        + 2.5 * np.log10(0.7 / fwhm_eff)
        + 1.25 * np.log10(exp_time / 30.0)
        - k_atm * (airmass - 1.0)
        - 1.1 * tau_cloud
    )
    if nexp > 1:
        m5 = 1.25 * np.log10(nexp * 10 ** (0.8 * m5))
    return m5


def m5_flat_sed(visit_filter, musky, fwhm_eff, exp_time, airmass, nexp=1, tau_cloud=0):
    """Calculate the m5 value, using photometric scaling.  Note, does not include shape of the object SED.

    Parameters
    ----------
    visit_filter : str
         One of u,g,r,i,z,y
    musky : float
        Surface brightness of the sky in mag/sq arcsec
    fwhm_eff : float
        The seeing effective FWHM (arcsec)
    exp_time : float
        Exposure time for each exposure in the visit.
    airmass : float
        Airmass of the observation (unitless)
    nexp : int, optional
        The number of exposures. Default 1.  (total on-sky time = exp_time * nexp)
    tau_cloud : float (0.)
        Any extinction from clouds in magnitudes (positive values = more extinction)

    Returns
    -------
    m5 : float
        The five-sigma limiting depth of a point source observed in the given conditions.
    """

    # Set up expected extinction (kAtm) and m5 normalization values (Cm) for each filter.
    # The Cm values must be changed when telescope and site parameters are updated.
    #
    # These values are calculated using $SYSENG_THROUGHPUTS/python/calcM5.py.
    # This set of values are calculated using v1.2 of the SYSENG_THROUGHPUTS repo.
    # The exposure time scaling depends on knowing the value of the exposure time used to calculate Cm/etc.

    # Only define the dicts once on initial call
    if not hasattr(m5_flat_sed, "Cm"):
        # Using Cm / d_cm_infinity values calculated for a 1x30s visit.
        # This results in an error of about 0.01 mag in u band for 2x15s visits (< in other bands)
        # See https://github.com/lsst-pst/survey_strategy/blob/master/fbs_1.3/m5FlatSed%20update.ipynb
        # for a more in-depth evaluation.
        sev = SysEngVals()

        m5_flat_sed.base_exp_time = sev.exptime
        m5_flat_sed.cm = sev.cm
        m5_flat_sed.d_cm_infinity = sev.d_cm_infinity
        m5_flat_sed.k_atm = sev.k_atm
        m5_flat_sed.msky = sev.sky_mag
    # Calculate adjustment if readnoise is significant for exposure time
    # (see overview paper, equation 7)
    tscale = (
        exp_time / m5_flat_sed.base_exp_time * np.power(10.0, -0.4 * (musky - m5_flat_sed.msky[visit_filter]))
    )
    d_cm = 0.0
    d_cm += m5_flat_sed.d_cm_infinity[visit_filter]
    d_cm -= 1.25 * np.log10(1 + (10 ** (0.8 * m5_flat_sed.d_cm_infinity[visit_filter]) - 1) / tscale)
    # Calculate m5 for 1 exp - 30s and other constants here come from definition of Cm/d_cm_infinity
    m5 = (
        m5_flat_sed.cm[visit_filter]
        + d_cm
        + 0.50 * (musky - 21.0)
        + 2.5 * np.log10(0.7 / fwhm_eff)
        + 1.25 * np.log10(exp_time / 30.0)
        - m5_flat_sed.k_atm[visit_filter] * (airmass - 1.0)
        - 1.1 * tau_cloud
    )
    # Then combine with coadd if >1 exposure
    if nexp > 1:
        m5 = 1.25 * np.log10(nexp * 10 ** (0.8 * m5))
    return m5



import unittest
import warnings

import healpy as hp
import numpy as np

import rubin_scheduler.skybrightness_pre as sbp

REMOTE_SKY_URL = (
    "https://s3df.slac.stanford.edu/groups/rubin/static/sim-data/sims_skybrightness_pre/h5_2023_09_12/"
)


class TestSkyPre(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        try:
            cls.sm = sbp.SkyModelPre(init_load_length=3, load_length=3)
            mjd = cls.sm.mjds[1] + 4.0 / 60.0 / 24.0
            tmp = cls.sm.return_mags(mjd)
            cls.nside = hp.npix2nside(tmp["r"].size)
            cls.data_present = True
        except ValueError:
            cls.data_present = False
            warnings.warn("Data files not found, skipping tests. Check data/ for instructions to pull data.")

    def test_return_mags(self):
        """
        Test all the ways ReturnMags can be used
        """
        timestep_max = 15.0 / 60.0 / 24.0
        # Check both the healpix and opsim fields
        if self.data_present:
            sms = [self.sm]
            mjds = []
            for mjd in sms[0].mjds[100:102]:
                mjds.append(mjd)
                mjds.append(mjd + 0.0002)

            # Make sure there's an mjd that is between sunrise/set that gets tested
            diff = sms[0].mjds[1:] - sms[0].mjds[0:-1]
            between = np.where(diff >= timestep_max)[0][0]
            mjds.append(sms[0].mjds[between + 1] + timestep_max)

            indxes = [None, [100, 101]]
            filters = [["u", "g", "r", "i", "z", "y"], ["r"]]

            for sm in sms:
                for mjd in mjds:
                    for indx in indxes:
                        for filt in filters:
                            mags = sm.return_mags(mjd, indx=indx, filters=filt)
                            # Check the filters returned are correct
                            self.assertEqual(len(filt), len(list(mags.keys())))
                            self.assertEqual(set(filt), set(mags.keys()))
                            # Check the magnitudes are correct
                            if indx is not None:
                                self.assertEqual(
                                    np.size(mags[list(mags.keys())[0]]),
                                    np.size(indx),
                                )

    def test_various(self):
        """
        Test some various loading things
        """
        # check that the sims_data stuff loads
        sm = sbp.SkyModelPre(init_load_length=3)
        mjd = self.sm.mjds[10] + 0.1
        mags = sm.return_mags(mjd)
        assert mags is not None

    @unittest.skip("Skipping remote sky test (too slow)")
    def test_remote_sky(self):
        sm = sbp.SkyModelPre(init_load_length=3, data_path=REMOTE_SKY_URL)
        min_mjd = np.min(sm.mjds)
        self.assertTrue(40000 < min_mjd < 80000)
        test_mjd = np.ceil(min_mjd) + 0.2
        mags = sm.return_mags(test_mjd)
        for band in mags.keys():
            npix = mags[band].shape[0]
            try:
                hp.npix2nside(npix)
                valid_npix = True
            except ValueError:
                valid_npix = False
            self.assertTrue(valid_npix)


if __name__ == "__main__":
    unittest.main()



__all__ = ("SeeingData", "ConstantSeeingData")

import os
import sqlite3
from dataclasses import dataclass

import numpy as np
from astropy.time import Time

from rubin_scheduler.data import get_data_dir


class SeeingData:
    """Read the seeing data from disk and return appropriate FWHM_500 value at a given time.
    This is for use in simulations only. Otherwise data would come from the EFD.

    Parameters
    ----------
    start_time : astropy.time.Time
        The time of the start of the simulation.
        The seeing database will be assumed to start on Jan 01 of the same year.
    seeing_db : str or None, optional
        The name of the seeing database.
        If None (default), this will use the simsee_pachon_58777_13.db file in the 'data' directory
        of this package.
        Other available seeing databases from sims_seeingModel include:
        seeing.db (the original, less-variable, 3 year seeing database)
        simsee_pachon_58777_13.db (the current default, 10 year, seeing database)
        simsee_pachon_58777_16.db (a similar, but slightly offset, 13 year seeing database)
        For more info on simsee_pachon_58777_*, see https://github.com/lsst/sims_seeingModel/issues/2
    offset_year : float, optional
        Offset into the cloud database by 'offset_year' years. Default 0.
    """

    def __init__(self, start_time, seeing_db=None, offset_year=0):
        self.seeing_db = seeing_db
        if self.seeing_db is None:
            self.seeing_db = os.path.join(get_data_dir(), "site_models", "simsee_pachon_58777_13.db")

        # Seeing database starts in Jan 01 of the year of the start of the simulation
        year_start = start_time.datetime.year + offset_year
        self.start_time = Time("%d-01-01" % year_start, format="isot", scale="tai")

        self.seeing_dates = None
        self.seeing_values = None
        self.read_data()

    def __call__(self, time):
        """Get the FWHM_500 value for the specified time.

        Parameters
        ----------
        time : astropy.time.Time
            Time in the simulation for which to find the 'current' zenith seeing values.
            The difference between this time and the start_time, plus the offset,
            will be used to query the seeing database.

        Returns
        -------
        float
            The FWHM_500(") closest to the specified time.
        """
        delta_time = (time - self.start_time).sec
        # Find the date to look for in the time range of the data.
        # Note that data dates should not necessarily start at zero.
        dbdate = delta_time % self.time_range + self.min_time
        idx = np.searchsorted(self.seeing_dates, dbdate)
        # searchsorted ensures that left < date < right
        # but we need to know if date is closer to left or to right
        left = self.seeing_dates[idx - 1]
        right = self.seeing_dates[idx]
        if np.size(idx) == 1:
            if dbdate - left < right - dbdate:
                idx -= 1
        else:
            d1 = dbdate - left
            d2 = right - dbdate
            to_sub = np.where(d1 < d2)
            idx[to_sub] -= 1
        return self.seeing_values[idx]

    def read_data(self):
        """Read the seeing information from disk.

        The default behavior is to use the module stored database. However, an
        alternate database file can be provided. The alternate database file needs to have a
        table called *Seeing* with the following columns:

        seeingId
            int : A unique index for each seeing entry.
        s_date
            int : The time (in seconds) from the start of the simulation, for the seeing observation.
        seeing
            float : The FWHM of the atmospheric PSF (in arcseconds) at zenith.
        """
        with sqlite3.connect(self.seeing_db) as conn:
            cur = conn.cursor()
            query = "select s_date, seeing from Seeing order by s_date;"
            cur.execute(query)
            results = np.array(cur.fetchall())
            self.seeing_dates = np.hsplit(results, 2)[0].flatten()
            self.seeing_values = np.hsplit(results, 2)[1].flatten()
            cur.close()
        # Make sure seeing dates are ordered appropriately (monotonically increasing).
        ordidx = self.seeing_dates.argsort()
        self.seeing_dates = self.seeing_dates[ordidx]
        self.seeing_values = self.seeing_values[ordidx]
        self.min_time = self.seeing_dates[0]
        self.max_time = self.seeing_dates[-1]
        self.time_range = self.max_time - self.min_time

    def config_info(self):
        """Report information about configuration of this data.

        Returns
        -------
        OrderedDict
        """
        config_info = {}
        config_info["Start time for db"] = self.start_time
        config_info["Seeing database"] = self.seeing_db
        return config_info


@dataclass
class ConstantSeeingData:
    fwhm_500: float = 0.7

    def __call__(self, time):
        """A constant FWHM_500 value

        Parameters
        ----------
        time : `astropy.time.Time`
            It principle the time for which the seeing is returned,
            in practice this argumnet is ignored, and included for
            compatibility.

        Returns
        -------
        fwhm_500 : `float`
            The FWHM at 500nm, in arcseconds.
        """
        return self.fwhm_500



__all__ = ("calc_season",)

import numpy as np


def calc_season(ra, time):
    """Calculate the 'season' in the survey for a series of ra/time values of an observation.
    Based only on the RA of the point on the sky, it calculates the 'season' based on when the sun
    passes through this RA (this marks the start of a 'season').

    Note that seasons should be calculated using the RA of a fixed point on the sky, such as
    the slice_point['ra'] if calculating season values for a series of opsim pointings on the sky.
    To convert to integer seasons, use np.floor(seasons)

    Parameters
    ----------
    ra : `float`
        The RA (in degrees) of the point on the sky
    time : `np.ndarray`
        The times of the observations, in MJD days

    Returns
    -------
    seasons : `np.array`
        The season values, as floats.
    """
    # A reference time and sun RA location to anchor the location of the Sun
    # This time was chosen as it is close to the expected start of the survey.
    ref_time = 60575.0
    ref_sun_ra = 179.20796047239727
    # Calculate the fraction of the sphere/"year" for this location
    offset = (ra - ref_sun_ra) / 360 * 365.25
    # Calculate when the seasons should begin
    season_began = ref_time + offset
    # Calculate the season value for each point.
    seasons = (time - season_began) / 365.25
    # (usually) Set first season at this point to 0
    seasons = seasons - np.floor(np.min(seasons))
    return seasons
    # The reference values can be evaluated using:
    # from astropy.time import Time
    # from astropy.coordinates import get_sun
    # from astropy.coordinates import EarthLocation
    # loc = EarthLocation.of_site('Cerro Pachon')
    # t = Time('2024-09-22T00:00:00.00', format='isot', scale='utc', location=loc)
    # print('Ref time', t.utc.mjd)
    # print('Ref sun RA', get_sun(t).ra.deg, t.utc.mjd)
    # print('local sidereal time at season start', t.sidereal_time('apparent').deg)



__all__ = ("gnomonic_project_toxy", "gnomonic_project_tosky")

import numpy as np


def gnomonic_project_toxy(ra1, dec1, r_acen, deccen):
    """Calculate x/y projection of ra1/dec1 in system with center at r_acen, deccen.
    Input radians. Grabbed from sims_selfcal"""
    # also used in Global Telescope Network website
    cosc = np.sin(deccen) * np.sin(dec1) + np.cos(deccen) * np.cos(dec1) * np.cos(ra1 - r_acen)
    x = np.cos(dec1) * np.sin(ra1 - r_acen) / cosc
    y = (np.cos(deccen) * np.sin(dec1) - np.sin(deccen) * np.cos(dec1) * np.cos(ra1 - r_acen)) / cosc
    return x, y


def gnomonic_project_tosky(x, y, r_acen, deccen):
    """Calculate RA/dec on sky of object with x/y and RA/Cen of field of view.
    Returns Ra/dec in radians."""
    denom = np.cos(deccen) - y * np.sin(deccen)
    RA = r_acen + np.arctan2(x, denom)
    dec = np.arctan2(np.sin(deccen) + y * np.cos(deccen), np.sqrt(x * x + denom * denom))
    return RA, dec



import glob
import os

import h5py
import numpy as np

from rubin_scheduler.data import get_data_dir

if __name__ == "__main__":
    # Gererate an sky map for each filters that is an estimate of how faint that part of sky
    # can get

    data_dir = get_data_dir()

    sky_files = glob.glob(os.path.join(data_dir, "skybrightness_pre/*.h5"))

    maximum_maps = {}
    filternames = "ugrizy"
    for filtername in filternames:
        maximum_maps[filtername] = []

    for skyfile in sky_files:
        h5 = h5py.File(skyfile, "r")
        for filtername in filternames:
            maximum_maps[filtername].append(np.nanmax(h5["sky_mags"][filtername], axis=0))
        h5.close()

    for filtername in filternames:
        maximum_maps[filtername] = np.nanmax(maximum_maps[filtername], axis=0)
    # convert dict to array
    dark_maps = np.empty(np.size(maximum_maps["r"]), dtype=list(zip(filternames, [float] * 6)))
    for filtername in filternames:
        dark_maps[filtername] = maximum_maps[filtername]
    np.savez(os.path.join(data_dir, "skybrightness_pre", "dark_maps.npz"), dark_maps=dark_maps)



import unittest

import numpy as np

import rubin_scheduler.scheduler.features as features
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.utils import empty_observation


class TestFeatures(unittest.TestCase):
    def test_pair_in_night(self):
        pin = features.PairInNight(gap_min=25.0, gap_max=45.0)
        self.assertEqual(np.max(pin.feature), 0.0)

        indx = np.array([1000])

        delta = 30.0 / 60.0 / 24.0

        # Add 1st observation, feature should still be zero
        obs = empty_observation()
        obs["filter"] = "r"
        obs["mjd"] = 59000.0
        pin.add_observation(obs, indx=indx)
        self.assertEqual(np.max(pin.feature), 0.0)

        # Add 2nd observation
        obs["mjd"] += delta
        pin.add_observation(obs, indx=indx)
        self.assertEqual(np.max(pin.feature), 1.0)

        obs["mjd"] += delta
        pin.add_observation(obs, indx=indx)
        self.assertEqual(np.max(pin.feature), 2.0)

    def test_conditions(self):
        observatory = ModelObservatory()
        conditions = observatory.return_conditions()
        self.assertIsInstance(repr(conditions), str)
        self.assertIsInstance(str(conditions), str)

        step_days = 1.0

        # Number of sidereal days in a standard day
        sidereal_hours_per_day = 24 * (24.0 / 23.9344696)
        initial_lmst = float(conditions.lmst)
        conditions.mjd = conditions.mjd + step_days
        new_lmst = float(conditions.lmst)
        self.assertAlmostEqual(
            new_lmst,
            (initial_lmst + step_days * sidereal_hours_per_day) % 24,
        )

        # Test that the string representation works
        _ = conditions.__str__()

        # Check that naked conditions work
        conditions_naked = features.Conditions()
        _ = conditions_naked.__str__()

    def test_note_last_observed(self):
        note_last_observed = features.NoteLastObserved(note="test")

        observation = empty_observation()
        observation["mjd"] = 59000.0

        note_last_observed.add_observation(observation=observation)

        assert note_last_observed.feature is None

        observation["note"] = "foo"

        note_last_observed.add_observation(observation=observation)
        assert note_last_observed.feature is None

        observation["note"] = "test"

        note_last_observed.add_observation(observation=observation)
        assert note_last_observed.feature == observation["mjd"]

    def test_note_last_observed_with_filter(self):
        note_last_observed = features.NoteLastObserved(
            note="test",
            filtername="r",
        )

        observation = empty_observation()
        observation["mjd"] = 59000.0

        note_last_observed.add_observation(observation=observation)

        assert note_last_observed.feature is None

        observation["note"] = "foo"

        note_last_observed.add_observation(observation=observation)
        assert note_last_observed.feature is None

        observation["note"] = "test"
        observation["filter"] = "g"

        note_last_observed.add_observation(observation=observation)
        assert note_last_observed.feature is None

        observation["note"] = "test"
        observation["filter"] = "r"

        note_last_observed.add_observation(observation=observation)
        assert note_last_observed.feature == observation["mjd"]


if __name__ == "__main__":
    unittest.main()



__all__ = ("CloudModel",)

import warnings

import numpy as np


class CloudModel:
    """LSST cloud calculations for cloud extinction.
    Currently this actually only returns the cloud coverage of the sky, exactly as reported in the
    cloud database (thus the sky coverage, in fractions of 8ths).


    Parameters
    ----------
    XXX--update docstring

    self.efd_requirements and self.map_requirements are also set.
    efd_requirements is a tuple: (list of str, float).
    This corresponds to the data columns required from the EFD and the amount of time history required.
    target_requirements is a list of str.
    This corresponds to the data columns required in the target map dictionary passed when calculating the
    processed telemetry values.
    """

    def __init__(self, cloud_column="cloud", altitude_column="altitude", azimuth_column="azimuth"):
        self.altcol = altitude_column
        self.azcol = azimuth_column
        self.cloudcol = cloud_column

    def configure(self, config=None):
        """"""
        warnings.warn("The configure method is deprecated.")

    def config_info(self):
        """"""
        warnings.warn("The config_info method is deprecated.")

    def __call__(self, cloud_value, altitude):
        """Calculate the sky coverage due to clouds.

        This is where we'd plug in Peter's cloud transparency maps and predictions.
        We could also try translating cloud transparency into a cloud extinction.
        For now, we're simply returning the cloud coverage that we already got from the database,
        but multiplied over the whole sky to provide a map.

        Parameters
        ----------
        cloud_value: float or efdData dict
            The value to give the clouds (XXX-units?).
        altitude:  float, np.array, or targetDict
            Altitude of the output (arbitrary).

        Returns
        -------
        dict of np.ndarray
            Cloud transparency map values.
        """
        if isinstance(cloud_value, dict):
            cloud_value = cloud_value[self.cloudcol]
        if isinstance(altitude, dict):
            altitude = altitude[self.altcol]

        model_cloud = np.zeros(len(altitude), float) + cloud_value
        return {"cloud": model_cloud}
